/*
 *			R E C E I V E. C
 * This program creates a AF_INET (InterNet Domain) raw soocket,
 * binds a name to it, and then reads from the socket.
 */

#include "vbrsr.h"

#define PROCPRIORITY -39	/* change priority of process */
#define C(x)	((x) & 0xff)

static struct
{
  bool compute_bit_err;			/* compute number of wrong bits */
  bool save_messages;			/* save messages */
  bool list;				/* list information */
  bool background;			/* process runs in the background */
  char pattern_type;			/* pattern type */
  u_int32_t pattern_value;		/* pattern value */
  double totNobit;			/* total No. of received bits */
  double totNoWbit;			/* total No. of wrong bits received */
  u_int32_t pkt_length;			/* length of received packet */
  u_int32_t pkt_counter;		/* number of received packet */
} g;					/* global structure */


char *ifname;				/* interface name */

static char *progname;			/* pointer to the program name */
static FILE *fp;			/* file pointer for the output */
static FILE *fperr;			/* file ptr for error_bit file */
static int sd;				/* socket descriptor */
static struct sockaddr_in from;		/* source address */
static char rcvbuf[1024];		/* receive area */
static struct iphdr iph;		/* Internet Protocol header */
static int nbyte = 0;


/*
 * Open raw socket on which to receive messages.
 *
 * input:  Host name.
 * output: none or error message.
 */
static bool
open_raw_socket (char *hostname)
{
  //  struct hostent *hp = NULL;

  /* create socket on which to receive */
  if ((sd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
    {
      perror("while opening socket");
      return FALSE;
    }

  /* get network address of the specified host.
  if ((hp = gethostbyname(hostname)) == NULL)
    {
      fprintf(stderr, "%s: %s\n", hostname, hstrerror(h_errno));
      return FALSE;
    } */

  /* create name with no wildcards
  from.sin_family = AF_INET;
  from.sin_port = 0;
  from.sin_addr = *(struct in_addr *)hp->h_addr; */

  /* associates a name with socket descriptor */
  if (bind(sd, (struct sockaddr *)&from, sizeof(from)) < 0)
    {
      perror("bind");
      return FALSE;
    }

  #if 0
    /* Source address. */
    printf("\n Source address\t");
    printf("%lu.%lu.%lu.%lu\n",
	   C(from.sin_addr.s_addr >> 24), C(from.sin_addr.s_addr >> 16),
	   C(from.sin_addr.s_addr >> 8), C(from.sin_addr.s_addr));
  #endif

  return TRUE;
} /* open_raw_socket() */


/*
 * Comute:
 * 	Total number of bit. Total number of wrong bit.
 *	Total number of received packet.
 *	(fixed, incremental and random pattern)
 * input:  none
 * output: none
 */
static void
compute_biterror_fixed ()
{
  register u_int32_t *data = (u_int32_t *)(rcvbuf + sizeof(struct iphdr));
  register u_int32_t pattern = htonl(g.pattern_value);
  register int NoWb = 0;
  register u_int8_t Wb;
  register int bitno;
  register int i, k;

  for (i = 0, k = 0, bitno = g.totNobit;
       i < g.pkt_length / sizeof(int);
       i++, bitno += sizeof(int))
    {
      #if 0
	printf("i=%d pattern=%u data=%u \n",i, pattern, *data);
      #endif
	if ((Wb = *data++ ^ pattern) != 0)
        do
	{
	  if (Wb & 0x01)
	    {
	      NoWb += 1;
	      /* record wrong bit */
	      if (g.list)
		fprintf(fperr, "%d\n", bitno + k);
	    }
	  k++;
	}
      while (Wb >>= 1);
    }
  /* save the total number of wrong bit */
  g.totNoWbit += NoWb;

} /* compute_biterror_fixed() */


static void
compute_biterror_incremental ()
{
  register u_int8_t *data = (u_int8_t *)(rcvbuf + sizeof(struct iphdr));
  register u_int32_t pattern = g.pattern_value;
  register int NoWb = 0;
  register u_int8_t Wb;
  register int bitno;
  register int i, k;

  for (i = 0, k = 0, bitno = g.totNobit;
       i < g.pkt_length;
       i++, bitno += 8)
    {
      pattern  = i + g.pattern_value;
      #if 0
	printf("i=%d pattern=%c data=%c \n",i, pattern, *data);
      #endif
	if ((Wb = *data++ ^ pattern) != 0)
        do
	{
	  if (Wb & 0x01)
	    {
	      NoWb += 1;
	      /* record wrong bit */
	      if (g.list)
		fprintf(fperr, "%d\n", bitno + k);
	    }
	  k++;
	}
      while (Wb >>= 1);
    }
  /* save the total number of wrong bit */
  g.totNoWbit += NoWb;

} /* compute_biterror_incremental() */


static void
compute_biterror_random ()
{
  register short *data = (short *)(rcvbuf + sizeof(struct iphdr));
  register u_int32_t pattern = g.pattern_value;
  register int NoWb = 0;
  register u_int8_t Wb;
  register int bitno;
  register int i, k;

  /* random starting point */
  srand(g.pattern_value);

  for (i = 0, k = 0, bitno = g.totNobit;
       i < g.pkt_length / sizeof(short);
       i++, bitno += sizeof(short))
    {
      pattern  = rand();
      #if 0
	printf("i=%d pattern=%c data=%c \n",i, pattern, *data);
      #endif
	if ((Wb = *data++ ^ pattern) != 0)
        do
	{
	  if (Wb & 0x01)
	    {
	      NoWb += 1;
	      /* record wrong bit */
	      if (g.list)
		fprintf(fperr, "%d\n", bitno + k);
	    }
	  k++;
	}
      while (Wb >>= 1);
    }
  /* save the total number of wrong bit */
  g.totNoWbit += NoWb;

} /* compute_biterror_random() */


/*
 * Catch SIGINT signal.
 */
static void
sigcatch (int sig)
{

  if (sig == SIGINT)
    {
      fprintf(stderr,"\n%s\t\n",progname);
      fprintf(stderr, "\tTotal number of received packet %u\n", g.pkt_counter);
      fprintf(stderr, "\tTotal number of bit %.1e\n", g.totNobit);
      fprintf(stderr, "\tTotal number of wrong bit %.1e\n", g.totNoWbit);
      fprintf(stderr, "\tBER = %.1e\n", g.totNoWbit / g.totNobit);

      signal(SIGINT, sigcatch);
    }
  return;
} /* sigcatch() */


static void
usage (void)
{
  fprintf(stderr, "usage: %s [options] host [ofile]\n\
	For a description of the options, try %s -h.\n", progname, progname);
  exit(2);
} /* usage() */


static void
help (void)
{
  fprintf(stderr, "\n\
\tusage: %s [options] hostname [ofile] [errfile]\n\
\n\
\n\
  where:<hostname> is the IP source addres, given by name or expressed \n\
	in the Internet dot notation.\n\
	<ofile> is the output file containing the data to be received\n\
	<errfile> is the output file containing the error bits\n\
\n\
	options are [defaults]:\n\
\n\
	-F|I|R <value> <type> means that, if no <dfile> is specified, the\n\
		packets must be filled with a pattern. <value> is one of:\n\
      		  F  packet filled with four bytes value [%08x]\n\
		  I  packet filled with value incremented each word [0]\n\
		  R  packet filled with random words with seed [time]\n\
		to enable or disable printing of information\n\
        -s	write the data to disk\n\
	-b	compute wrong bits\n\
	-l	record wrong bits\n\
	-h	help: print this page.\n",
	   progname, default_pattern);
  exit(2);
} /* help() */

int
main (int argc, char **argv)
{
  char c;
  extern char *optarg;
  extern int optind;
  char *hostname = NULL;
  char *out_file = NULL;
  char *errlist = NULL;
  int child;
  struct sched_param prio;


  /* initialize variables */
  g.compute_bit_err = FALSE;
  g.save_messages = FALSE;
  g.background = FALSE;
  g.list = FALSE;
  g.pkt_counter = 0;
  g.totNoWbit = 0;
  g.totNobit = 0;


  /* create a child process */
  if ((child = fork()) < 0)
    perrexit2("fork");
  if (child > 0)
    {
      fprintf(stderr, "%s: Running in the background. Process %d\n",
	      argv [0], child);
      exit(1);
    }

  /* program name pointer */
  progname = argv[0];

  /* catch interrupt signal */
  signal(SIGINT, sigcatch);

  while ((c = getopt(argc, argv, "F::I::R::slbh")) != -1)
    switch (c)
      {
      case FIXED:
	if (optarg == NULL && g.pattern_value == -1)
	  g.pattern_value = default_pattern;
      case INCREMENTAL:
	if (optarg == NULL && g.pattern_value == -1)
	  g.pattern_value = 0;
      case RANDOM:
	if (optarg == NULL && g.pattern_value == -1)
	  g.pattern_value = time(NULL);
	if (optarg != NULL)
	  if (sscanf(optarg, "%i", &g.pattern_value) != 1
	      || g.pattern_value <= 0)
	    {
	      fprintf(stderr, "argument to option -%c must be positive\n", c);
	      usage();
	    }
	g.pattern_type = c;
	break;

      case 'b':
	g.compute_bit_err = TRUE;
	break;

      case 'l':
	g.list = TRUE;
	break;

      case 's':
        g.save_messages = TRUE;
	break;

      case 'h':
	help();
	break;
      case '?':
	usage();
	break;
      } /* switch */

  /* Get host name */
  if (optind == argc)
    {
      fprintf(stderr, "hostname must be specified\n");
      usage();
    }
  else
    hostname = argv[optind];

  /* Get output file name */
  if (g.save_messages == TRUE)
    {
      if (optind+1 == argc)
        {
          fprintf(stderr, "output file must be specified\n");
          usage();
        }
      else
        out_file = argv[optind+1];
    }

  /* Get `errlist' file name */
  if (g.list)
    {
      if (optind+2 == argc)
        {
          fprintf(stderr,"errlist file must be specified\n");
          usage();
        }
      else
        errlist = argv[optind+2];
    }

  /* Get host name */
  if (!open_raw_socket(hostname))
    usage();

  /* Open output file name */
  if (g.save_messages == TRUE)
    {
      if ((fp = fopen(out_file, "w+")) == NULL)
        perrexit2("while opening output file");
    }

  /* Open `errlist' file name */
  if (g.list)
    {
      if ((fperr = fopen(errlist, "w+")) == NULL)
	perrexit2("while opening errlist file");
    }

  /* clear receive area */
  bzero(rcvbuf, sizeof(rcvbuf));

  /* Set real-time scheduling priority */
  prio.sched_priority =
    (sched_get_priority_min(POLICY) + sched_get_priority_max(POLICY)) / 2;
  if ((sched_setscheduler(0, POLICY, &prio) < 0))
    {
      perror("sched_setscheduler");
      usage();
    }

  /* read the data from the socket in the loop */
  do
    {
      /* read the data from the socket */
      if ((nbyte = read(sd, rcvbuf, sizeof(rcvbuf))) < 0)
        {
	  if (errno == EINTR)
	    continue;
	  perror("while reading socket");
	  exit(2);
	}

      /* increment packet counter */
      g.pkt_counter += 1;

      /* compute number of wrong bit */
      if (g.compute_bit_err)
	{
          /* copy Internet header */
          (void) bzero((char *)&iph, sizeof(struct iphdr));
          bcopy(rcvbuf, (char *)&iph, sizeof(struct iphdr));

          #if 0
          /* Construct an Internet address representation. */
          printf("\n Source address\t");
          printf("%lu.%lu.%lu.%lu\n", C(iph.i_shost >> 24), C(iph.i_shost>> 16),
		 C(iph.i_shost >> 8), C(iph.i_shost));
          printf("\n Destination address\t");
          printf("%lu.%lu.%lu.%lu\n", C(iph.i_dhost >> 24), C(iph.i_dhost>> 16),
		 C(iph.i_dhost >> 8), C(iph.i_dhost));

          printf(" Data length %d tototal_pkt_length %d\n",
		 iph.i_tlen - sizeof(struct iphdr), iph.i_tlen);
          #endif

          /* get length of packet */
          g.pkt_length = iph.tot_len - sizeof(struct iphdr);


          /* compute number of wrong bits */
          switch (g.pattern_type)
            {
              case FIXED:		/* fixed pattern */
		compute_biterror_fixed();
              break;
              case INCREMENTAL:	/* incremental pattern */
		compute_biterror_incremental();
              break;
              case RANDOM:		/* random pattern */
		compute_biterror_random();
              break;
            } /* switch(pattern) */

          /* increment number of total received bits */
          g.totNobit += 8 * g.pkt_length;
	} /* compute_bit_error */


      /* write the data to disk */
      if (g.save_messages)
	{
	  /* get current data length */
	  u_int32_t data_length = nbyte - sizeof(struct iphdr);

	  write_again:
	  if (fwrite(rcvbuf + sizeof(struct iphdr), data_length , 1, fp) != 1)
	    {
	      if (errno == EINTR)
		goto write_again;
	      perror("while writing output file");
	      exit(1);
	    }
	  #if 0
          printf("Data length %u data %s\n",
		 data_length,rcvbuf+sizeof(struct iphdr));
	  #endif
	} /* save messages */

    } while (TRUE);
    /* NOTREACHED */
} /* main() */



/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE"  "\\sw+_t" "bool")
 * End:
 */
